@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 1998-2016 Sergey Poznyakoff
+   Copyright (C) 1998-2018 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@ _add_load_dir (void *item, void *unused)
 }
 
 void
-dicod_loader_init()
+dicod_loader_init(void)
 {
     lt_dlinit();
     dico_list_iterate(prepend_load_path, _add_load_dir, NULL);
@@ -35,37 +35,16 @@ dicod_loader_init()
 
 #define MODULE_ASSERT(cond)					\
     if (!(cond)) {						\
-	lt_dlclose(handle);					\
 	dico_log(L_ERR, 0, _("%s: faulty module: (%s) failed"), \
 		 argv[0],					\
-	         #cond);					\
+		 #cond);					\
 	return 1;						\
     }
 
 static int
-dicod_load_module0(dicod_module_instance_t *inst, int argc, char **argv)
+module_init(dicod_module_instance_t *inst, struct dico_database_module *pmod,
+	    int argc, char **argv)
 {
-    lt_dlhandle handle = NULL;
-    lt_dladvise advise = NULL;
-    struct dico_database_module *pmod;    
-
-    if (inst->handle) {
-	dico_log(L_ERR, 0, _("module %s already loaded"), argv[0]);
-	return 1;
-    }
-	
-    if (!lt_dladvise_init(&advise) && !lt_dladvise_ext(&advise)
-        && !lt_dladvise_global(&advise))
-	handle = lt_dlopenadvise(argv[0], advise);
-    lt_dladvise_destroy(&advise);
-
-    if (!handle) {
-	dico_log(L_ERR, 0, _("cannot load module %s: %s"), argv[0],
-		 lt_dlerror());
-	return 1;
-    }
-    
-    pmod = (struct dico_database_module *) lt_dlsym(handle, "module");
     MODULE_ASSERT(pmod);
     MODULE_ASSERT(pmod->dico_version <= DICO_MODULE_VERSION);
     if (pmod->dico_capabilities & DICO_CAPA_NODB) {
@@ -78,148 +57,97 @@ dicod_load_module0(dicod_module_instance_t *inst, int argc, char **argv)
 	MODULE_ASSERT(pmod->dico_output_result);
 	MODULE_ASSERT(pmod->dico_result_count);
 	MODULE_ASSERT(pmod->dico_free_result);
-	
-	if (pmod->dico_open || pmod->dico_close)
-	    MODULE_ASSERT(pmod->dico_open && pmod->dico_close);
     }
-    
+
     if (pmod->dico_init && pmod->dico_init(argc, argv)) {
-	lt_dlclose(handle);
 	dico_log(L_ERR, 0, _("%s: initialization failed"), argv[0]);
 	return 1;
     }
 
-    inst->handle = handle;
     inst->module = pmod;
     return 0;
 }
 
+static int
+module_load(dicod_module_instance_t *inst, int argc, char **argv)
+{
+    lt_dlhandle handle = NULL;
+    lt_dladvise advise = NULL;
+    struct dico_database_module *pmod;
+
+    if (inst->handle) {
+	dico_log(L_ERR, 0, _("module %s already loaded"), argv[0]);
+	return 1;
+    }
+    if (inst->module)
+	/* FIXME: Built-in module */
+	return 0;
+
+    if (!lt_dladvise_init(&advise) && !lt_dladvise_ext(&advise)
+	&& !lt_dladvise_global(&advise))
+	handle = lt_dlopenadvise(argv[0], advise);
+    lt_dladvise_destroy(&advise);
+
+    if (!handle) {
+	dico_log(L_ERR, 0, _("cannot load module %s: %s"), argv[0],
+		 lt_dlerror());
+	return 1;
+    }
+
+    pmod = (struct dico_database_module *) lt_dlsym(handle, "module");
+
+    if (module_init(inst, pmod, argc, argv)) {
+	lt_dlclose(handle);
+	return 1;
+    }
+
+    inst->handle = handle;
+
+    return 0;
+}
+
 int
 dicod_load_module(dicod_module_instance_t *inst)
 {
     struct wordsplit ws;
     int rc;
-    
+
     if (wordsplit(inst->command, &ws, WRDSF_DEFFLAGS)) {
 	dico_log(L_ERR, 0, _("cannot parse command line `%s': %s"),
 		 inst->command, wordsplit_strerror (&ws));
 	return 1;
     }
-
-    rc = dicod_load_module0(inst, ws.ws_wordc, ws.ws_wordv);
-
+    rc = module_load(inst, ws.ws_wordc, ws.ws_wordv);
     wordsplit_free(&ws);
-    
     return rc;
 }
-
-int
-dicod_init_database(dicod_database_t *dp)
+
+static void
+builtin_module_load(char const *ident, struct dico_database_module *pmod)
 {
-    dicod_module_instance_t *inst = dp->instance;
-
-    if (inst->module->dico_capabilities & DICO_CAPA_NODB) {
-	dico_log(L_ERR, 0, _("cannot initialize database `%s': "
-			     "module `%s' does not support databases"),
-		 dp->command, inst->ident);
-	return 1;
+    dicod_module_instance_t *inst = xzalloc(sizeof(*inst));
+    char *argv[2];
+    inst->ident = xstrdup(ident);
+    inst->command = xstrdup(ident);
+    argv[0] = inst->command;
+    argv[1] = NULL;
+    if (module_init(inst, pmod, 1, argv)) {
+	dico_log(L_CRIT, 0,
+		 "INTERNAL ERROR: failed to load built-in module %s",
+		 ident);
+	dico_log(L_CRIT, 0, "Please, report");
+	abort();
     }
-    
-    if (inst->module->dico_init_db) {
-	dp->mod_handle = inst->module->dico_init_db(dp->name,
-						    dp->argc, dp->argv);
-	if (!dp->mod_handle) {
-	    dico_log(L_ERR, 0, _("cannot initialize database `%s'"),
-		     dp->command);
-	    return 1;
-	}
-    }
-    return 0;
-}
-
-int
-dicod_open_database(dicod_database_t *dp)
-{
-    dicod_module_instance_t *inst = dp->instance;
-
-    if (inst->module->dico_open) {
-	if (inst->module->dico_open(dp->mod_handle)) {
-	    dico_log(L_ERR, 0, _("cannot open database `%s'"),
-		     dp->command);
-	    return 1;
-	}
-    }
-    return 0;
-}
-
-int
-dicod_close_database(dicod_database_t *dp)
-{
-    int rc = 0;
-    
-    if (dp->mod_handle) {
-	dicod_module_instance_t *inst = dp->instance;
-	if (inst->module->dico_close) 
-	    rc = inst->module->dico_close(dp->mod_handle);
-    }
-    return rc;
-}
-
-int
-dicod_free_database(dicod_database_t *dp)
-{
-    int rc = 0;
-    
-    if (dp->mod_handle) {
-	dicod_module_instance_t *inst = dp->instance;
-	if (inst->module->dico_free_db) {
-	    rc = inst->module->dico_free_db(dp->mod_handle);
-	    dp->mod_handle = NULL;
-	}
-    }
-    return rc;
-}
-
-char *
-dicod_get_database_descr(dicod_database_t *db)
-{
-    if (db->descr)
-	return db->descr;
-    else {
-	dicod_module_instance_t *inst = db->instance;
-	if (inst->module->dico_db_descr)
-	    return inst->module->dico_db_descr(db->mod_handle);
-    }
-    return NULL;
+    inst->handle = NULL;
+    xdico_list_append(modinst_list, inst);
 }
 
 void
-dicod_free_database_descr(dicod_database_t *db, char *descr)
+dicod_builtin_module_init(void)
 {
-    if (descr && descr != db->descr)
-	free(descr);
+    builtin_module_load("virtual", &virtual_builtin_module);
 }
-
-char *
-dicod_get_database_info(dicod_database_t *db)
-{
-    if (db->info)
-	return db->info;
-    else {
-	dicod_module_instance_t *inst = db->instance;
-	if (inst->module->dico_db_info)
-	    return inst->module->dico_db_info(db->mod_handle);
-    }
-    return NULL;
-}
-
-void
-dicod_free_database_info(dicod_database_t *db, char *info)
-{
-    if (info && info != db->info)
-	free(info);
-}
-
+
 static int
 _dup_lang_item(void *item, void *data)
 {
@@ -246,32 +174,6 @@ dicod_any_lang_list_p(dico_list_t list)
 	   || (dico_list_count(list) == 1
 	       && strcmp (dico_list_item(list, 0), "*") == 0);
 }
-
-void
-dicod_get_database_languages(dicod_database_t *db, dico_list_t dlist[])
-{
-    if (!(db->flags & DICOD_DBF_LANG)) {
-	dicod_module_instance_t *inst = db->instance;
-	if (inst->module->dico_db_lang) {
-	    /* FIXME: Return code? */
-	    inst->module->dico_db_lang(db->mod_handle, db->langlist);
-	    if (db->langlist[0] || db->langlist[1]) {		
-		if (!db->langlist[0])
-		    db->langlist[0] = dicod_langlist_copy(db->langlist[1]);
-		else if (!db->langlist[1])
-		    db->langlist[1] = dicod_langlist_copy(db->langlist[0]);
-	    }
-	    if (dicod_any_lang_list_p(db->langlist[0]))
-		dico_list_destroy(&db->langlist[0]);
-	    if (dicod_any_lang_list_p(db->langlist[1]))
-		dico_list_destroy(&db->langlist[1]);
-	}
-	db->flags |= DICOD_DBF_LANG;
-    }
-    dlist[0] = db->langlist[0];
-    dlist[1] = db->langlist[1];
-}
-
 
 static char nomatch[] = "552 No match";
 static size_t nomatch_len = (sizeof(nomatch)-1);
@@ -301,34 +203,32 @@ dicod_word_first(dico_stream_t stream, const char *word,
 
     itr = xdico_list_iterator(database_list);
     for (db = dico_iterator_first(itr); db; db = dico_iterator_next(itr)) {
-	if (database_visible_p(db)) {
-	    struct dico_database_module *mp = db->instance->module;
-	    dico_result_t res = strat ?
-		            mp->dico_match(db->mod_handle, strat, word) :
-		            mp->dico_define(db->mod_handle, word);
+	if (database_is_visible(db)) {
+	    dico_result_t res = strat
+		? dicod_database_match(db, strat, word)
+		: dicod_database_define(db, word);
 	    size_t count;
-	    
+
 	    if (!res)
 		continue;
-	    count = mp->dico_result_count(res);
-	    
+	    count = dicod_database_result_count(db, res);
+
 	    if (count) {
 		if (strat)
 		    current_stat.matches = count;
 		else
 		    current_stat.defines = count;
-		if (mp->dico_compare_count)
-		    current_stat.compares = mp->dico_compare_count(res);
+		current_stat.compares = dicod_database_compare_count(db, res);
 		stream_printf(stream, begfmt, (unsigned long) count);
 		proc(db, res, word, stream, data, count);
 		stream_writez(stream, (char*) endmsg);
 		report_current_timing(stream, tid);
 		dico_stream_write(stream, "\n", 1);
 		access_log_status(begfmt, endmsg);
-		mp->dico_free_result(res);
+		dicod_database_result_free(db, res);
 		break;
 	    } else
-		mp->dico_free_result(res);
+		dicod_database_result_free(db, res);
 	}
     }
     dico_iterator_destroy(&itr);
@@ -355,7 +255,7 @@ dicod_word_all(dico_stream_t stream, const char *word,
     dico_list_t reslist = xdico_list_create();
     size_t total = 0;
     struct dbres *rp;
-    
+
     begin_timing(tid);
 
     if (strat && stratcl_check_word(strat->stratcl, word)) {
@@ -366,24 +266,22 @@ dicod_word_all(dico_stream_t stream, const char *word,
 
     itr = xdico_list_iterator(database_list);
     for (db = dico_iterator_first(itr); db; db = dico_iterator_next(itr)) {
-	if (database_visible_p(db)) {
-	    struct dico_database_module *mp = db->instance->module;
-	    dico_result_t res = strat ?
-		              mp->dico_match(db->mod_handle, strat, word) :
-		              mp->dico_define(db->mod_handle, word);
+	if (database_is_visible(db)) {
+	    dico_result_t res = strat
+		? dicod_database_match(db, strat, word)
+		: dicod_database_define(db, word);
 	    size_t count;
-	
+
 	    if (!res)
 		continue;
-	    count = mp->dico_result_count(res);
+	    count = dicod_database_result_count(db, res);
 	    if (!count) {
-		mp->dico_free_result(res);
+		dicod_database_result_free(db, res);
 		continue;
 	    }
 
 	    total += count;
-	    if (mp->dico_compare_count)
-		current_stat.compares += mp->dico_compare_count(res);
+	    current_stat.compares += dicod_database_compare_count(db, res);
 	    rp = xmalloc(sizeof(*rp));
 	    rp->db = db;
 	    rp->res = res;
@@ -399,7 +297,7 @@ dicod_word_all(dico_stream_t stream, const char *word,
 	dico_stream_writeln(stream, nomatch, nomatch_len);
     } else {
 	itr = xdico_list_iterator(reslist);
-	
+
 	if (strat)
 	    current_stat.matches = total;
 	else
@@ -407,7 +305,7 @@ dicod_word_all(dico_stream_t stream, const char *word,
 	stream_printf(stream, begfmt, (unsigned long) total);
 	for (rp = dico_iterator_first(itr); rp; rp = dico_iterator_next(itr)) {
 	    proc(rp->db, rp->res, word, stream, data, rp->count);
-	    rp->db->instance->module->dico_free_result(rp->res);
+	    dicod_database_result_free(rp->db, rp->res);
 	    free(rp);
 	}
 	stream_writez(stream, (char*) endmsg);
@@ -425,13 +323,12 @@ print_matches(dicod_database_t *db, dico_result_t res,
 	      dico_stream_t stream, void *data, size_t count)
 {
     size_t i;
-    struct dico_database_module *mp = db->instance->module;
     dico_stream_t ostr = data;
-    
+
     for (i = 0; i < count; i++) {
 	stream_writez(ostr, db->name);
 	dico_stream_write(ostr, " \"", 2);
-	mp->dico_output_result(res, i, ostr);
+	dicod_database_result_output(db, res, i, ostr);
 	dico_stream_write(ostr, "\"\n", 2);
     }
 }
@@ -440,29 +337,28 @@ void
 dicod_match_word_db(dicod_database_t *db, dico_stream_t stream,
 		    const dico_strategy_t strat, const char *word)
 {
-    struct dico_database_module *mp = db->instance->module;
     dico_result_t res;
     size_t count;
-    
+
     begin_timing("match");
-    res = mp->dico_match(db->mod_handle, strat, word);
-    
+
+    res = dicod_database_match(db, strat, word);
+
     if (!res) {
 	access_log_status(nomatch, nomatch);
 	dico_stream_writeln(stream, nomatch, nomatch_len);
 	return;
     }
 
-    count = mp->dico_result_count(res);
+    count = dicod_database_result_count(db, res);
     if (count == 0) {
 	access_log_status(nomatch, nomatch);
 	dico_stream_writeln(stream, nomatch, nomatch_len);
     } else {
 	dico_stream_t ostr;
-	
+
 	current_stat.matches = count;
-	if (mp->dico_compare_count)
-	    current_stat.compares = mp->dico_compare_count(res);
+	current_stat.compares = dicod_database_compare_count(db, res);
 	stream_printf(stream, "152 %lu matches found: list follows\n",
 		      (unsigned long) count);
 	ostr = dicod_ostream_create(stream, NULL);
@@ -476,8 +372,8 @@ dicod_match_word_db(dicod_database_t *db, dico_stream_t stream,
 	dico_stream_write(stream, "\n", 1);
 	access_log_status("152", "250");
     }
-    
-    mp->dico_free_result(res);
+
+    dicod_database_result_free(db, res);
 }
 
 void
@@ -507,8 +403,6 @@ dicod_match_word_all(dico_stream_t stream,
     dico_stream_close(ostr);
     dico_stream_destroy(&ostr);
 }
-
-
 
 static void
 print_definitions(dicod_database_t *db, dico_result_t res,
@@ -516,58 +410,50 @@ print_definitions(dicod_database_t *db, dico_result_t res,
 		  dico_stream_t stream, void *data, size_t count)
 {
     size_t i;
-    char *descr = dicod_get_database_descr(db);
-    struct dico_database_module *mp = db->instance->module;
+    char *descr = dicod_database_get_descr(db);
+
     for (i = 0; i < count; i++) {
 	dico_stream_t ostr;
-	dico_assoc_list_t hdr = NULL;
-	
+	dico_assoc_list_t hdr;
+
 	stream_printf(stream, "151 \"%s\" %s \"%s\"\n",
 		      word, db->name, descr ? descr : "");
-	if (mp->dico_result_headers) {
-	    if (db->mime_headers)
-		hdr = dico_assoc_dup(db->mime_headers);
-	    else
-		dico_header_parse(&hdr, NULL);
-	    mp->dico_result_headers(res, hdr);
-	    ostr = dicod_ostream_create(stream, hdr);
-	} else
-	    ostr = dicod_ostream_create(stream, db->mime_headers);
-	mp->dico_output_result(res, i, ostr);
+
+	hdr = dicod_database_mime_header(db, res);
+	ostr = dicod_ostream_create(stream, hdr);
+	dicod_database_result_output(db, res, i, ostr);
 	total_bytes_out += dico_stream_bytes_out(ostr);
 	dico_stream_close(ostr);
 	dico_stream_destroy(&ostr);
 	dico_stream_write(stream, "\n.\n", 3);
 	dico_assoc_destroy(&hdr);
     }
-    dicod_free_database_descr(db, descr);
+    dicod_database_free_descr(db, descr);
 }
 
 void
 dicod_define_word_db(dicod_database_t *db, dico_stream_t stream,
 		     const char *word)
 {
-    struct dico_database_module *mp = db->instance->module;
     dico_result_t res;
     size_t count;
-    
+
     begin_timing("define");
 
-    res = mp->dico_define(db->mod_handle, word);
+    res = dicod_database_define(db, word);
     if (!res) {
 	access_log_status(nomatch, nomatch);
 	dico_stream_writeln(stream, nomatch, nomatch_len);
 	return;
     }
 
-    count = mp->dico_result_count(res);
+    count = dicod_database_result_count(db, res);
     if (count == 0) {
 	access_log_status(nomatch, nomatch);
 	dico_stream_writeln(stream, nomatch, nomatch_len);
     } else {
 	current_stat.defines = count;
-	if (mp->dico_compare_count)
-	    current_stat.compares = mp->dico_compare_count(res);
+	current_stat.compares = dicod_database_compare_count(db, res);
 	stream_printf(stream, "150 %lu definitions found: list follows\n",
 		      (unsigned long) count);
 	print_definitions(db, res, word, stream, NULL, count);
@@ -576,8 +462,8 @@ dicod_define_word_db(dicod_database_t *db, dico_stream_t stream,
 	dico_stream_write(stream, "\n", 1);
 	access_log_status("150", "250");
     }
-    
-    mp->dico_free_result(res);
+
+    dicod_database_result_free(db, res);
 }
 
 void
@@ -603,8 +489,8 @@ dicod_module_test(int argc, char **argv)
 {
     int i;
     dicod_module_instance_t inst;
-    struct dico_database_module *pmod;    
-    
+    struct dico_database_module *pmod;
+
     /* Split arguments in two parts: unit test arguments and
        optional module initialization arguments */
     for (i = 0; i < argc; i++)
@@ -614,18 +500,18 @@ dicod_module_test(int argc, char **argv)
 
     if (i == 0)
 	dico_die(EX_UNAVAILABLE, L_ERR, 0, _("no module name"));
-    
+
     memset(&inst, 0, sizeof(inst));
     if (i < argc) {
 	argv[i] = argv[0];
-	if (dicod_load_module0(&inst, argc - i, argv + i))
+	if (module_load(&inst, argc - i, argv + i))
 	    return 1;
 	argv[i] = NULL;
     } else {
 	char *null_argv[2];
 	null_argv[0] = argv[0];
 	null_argv[1] = NULL;
-	if (dicod_load_module0(&inst, 1, null_argv))
+	if (module_load(&inst, 1, null_argv))
 	    return 1;
     }
 
