@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 2008, 2010, 2012 Sergey Poznyakoff
+   Copyright (C) 2008-2018 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -47,28 +47,28 @@ apply_catch_body(void *data)
 }
 
 static SCM
-eval_catch_handler (void *data, SCM tag, SCM throw_args)
+eval_catch_handler(void *data, SCM tag, SCM throw_args)
 {
     scm_handle_by_message_noexit("dico", tag, throw_args);
     longjmp(*(jmp_buf*)data, 1);
 }
 
 struct scheme_exec_data {
-    SCM (*handler) (void *data);
+    SCM (*handler)(void *data);
     void *data;
     SCM result;
 };
 
 static SCM
-scheme_safe_exec_body (void *data)
+scheme_safe_exec_body(void *data)
 {
     struct scheme_exec_data *ed = data;
-    ed->result = ed->handler (ed->data);
+    ed->result = ed->handler(ed->data);
     return SCM_BOOL_F;
 }
 
 static int
-guile_safe_exec(SCM (*handler) (void *data), void *data, SCM *result)
+guile_safe_exec(SCM (*handler)(void *data), void *data, SCM *result)
 {
     jmp_buf jmp_env;
     struct scheme_exec_data ed;
@@ -77,9 +77,9 @@ guile_safe_exec(SCM (*handler) (void *data), void *data, SCM *result)
 	return 1;
     ed.handler = handler;
     ed.data = data;
-    scm_internal_lazy_catch(SCM_BOOL_T,
-			    scheme_safe_exec_body, (void*)&ed,
-			    eval_catch_handler, &jmp_env);
+    scm_c_with_throw_handler(SCM_BOOL_T,
+			     scheme_safe_exec_body, (void*)&ed,
+			     eval_catch_handler, &jmp_env, 0);
     if (result)
 	*result = ed.result;
     return 0;
@@ -190,9 +190,9 @@ guile_call_proc(SCM *result, SCM proc, SCM arglist)
     }
     adata.proc = proc;
     adata.arg = arglist;
-    *result = scm_internal_lazy_catch(SCM_BOOL_T,
-				      apply_catch_body, &adata,
-				      eval_catch_handler, &jmp_env);
+    *result = scm_c_with_throw_handler(SCM_BOOL_T,
+				       apply_catch_body, &adata,
+				       eval_catch_handler, &jmp_env, 0);
     return 0;
 }
 
@@ -209,7 +209,7 @@ dico_new_scm_key(struct dico_key **pkey)
     SCM_RETURN_NEWSMOB(_guile_dico_key_tag, kptr);
 }
 
-static scm_sizet
+static size_t
 _guile_dico_key_free(SCM message_smob)
 {
     struct dico_key *kp = (struct dico_key *) SCM_CDR (message_smob);
@@ -231,7 +231,7 @@ _guile_dico_key_print(SCM message_smob, SCM port, scm_print_state *pstate)
 }
 
 static void
-_guile_init_dico_key()
+_guile_init_dico_key(void)
 {
     _guile_dico_key_tag =
 	scm_make_smob_type("Dico key", sizeof (struct dico_key));
@@ -283,7 +283,7 @@ _make_strategy(const dico_strategy_t strat)
     SCM_RETURN_NEWSMOB(_guile_strategy_tag, sp);
 }
 
-static scm_sizet
+static size_t
 _guile_strategy_free(SCM message_smob)
 {
     struct _guile_strategy *sp =
@@ -296,7 +296,7 @@ static int
 _guile_strategy_print(SCM message_smob, SCM port, scm_print_state * pstate)
 {
     struct _guile_strategy *sp =
-	(struct _guile_strategy *) SCM_CDR (message_smob);
+	(struct _guile_strategy *) SCM_CDR(message_smob);
     scm_puts("#<strategy ", port);
     scm_puts(sp->strat->name, port);
     scm_puts(" [", port);
@@ -306,7 +306,7 @@ _guile_strategy_print(SCM message_smob, SCM port, scm_print_state * pstate)
 }
 
 static void
-_guile_init_strategy()
+_guile_init_strategy(void)
 {
     _guile_strategy_tag = scm_make_smob_type("strategy",
 					     sizeof (struct _guile_strategy));
@@ -346,30 +346,30 @@ SCM_DEFINE_PUBLIC(scm_dico_strat_select_p, "dico-strat-select?", 3, 0, 0,
     sp = (struct _guile_strategy *) SCM_CDR(strat);
     stratp = sp->strat;
     
+    wordstr = scm_to_locale_string(word);
     if (scm_is_string(key)) {
 	char *keystr = scm_to_locale_string(key);
 	struct dico_key skey;
 
-	if (dico_key_init(&skey, stratp, wordstr)) {
-	    free(keystr);
+	rc = dico_key_init(&skey, stratp, keystr);
+	free(keystr);
+	if (rc) {
+	    free(wordstr);
 	    scm_misc_error(FUNC_NAME,
 			   "key initialization failed: ~S",
 			   scm_list_1(key));
 	}
 	rc = dico_key_match(&skey, wordstr);
 	dico_key_deinit(&skey);
-	free(wordstr);
-	free(keystr);
     } else {
 	struct dico_key *kptr;
 
 	SCM_ASSERT(CELL_IS_KEY(key), key, SCM_ARG3, FUNC_NAME);
     
 	kptr = (struct dico_key *) SCM_CDR(key);
-	wordstr = scm_to_locale_string(word);
 	rc = dico_key_match(kptr, wordstr);
-	free(wordstr);
     }
+    free(wordstr);
     return rc ? SCM_BOOL_T : SCM_BOOL_F;
 }
 #undef FUNC_NAME
@@ -502,7 +502,8 @@ SCM_DEFINE_PUBLIC(scm_dico_register_markup, "dico-register-markup", 1, 0, 0,
 	break;
 
     case ENOMEM:
-	scm_memory_error(FUNC_NAME);
+	scm_report_out_of_memory ();
+	break;
 
     case EINVAL:
 	scm_misc_error(FUNC_NAME,
@@ -528,7 +529,7 @@ SCM_DEFINE_PUBLIC(scm_dico_current_markup, "dico-current-markup", 0, 0, 0,
 #undef FUNC_NAME
 
 
-static scm_t_bits scm_tc16_dico_port;
+static scm_t_port_type *scm_dico_port_type;
 struct _guile_dico_port {
     dico_stream_t str;
 };
@@ -537,68 +538,31 @@ static SCM
 _make_dico_port(dico_stream_t str)
 {
     struct _guile_dico_port *dp;
-    SCM port;
-    scm_port *pt;
 
-    dp = scm_gc_malloc (sizeof (struct _guile_dico_port), "dico-port");
+    dp = scm_gc_typed_calloc (struct _guile_dico_port);
     dp->str = str;
-
-    port = scm_new_port_table_entry(scm_tc16_dico_port);
-    pt = SCM_PTAB_ENTRY(port);
-    pt->rw_random = 0;
-    SCM_SET_CELL_TYPE(port,
-		      (scm_tc16_dico_port | SCM_OPN | SCM_WRTNG | SCM_BUF0));
-    SCM_SETSTREAM(port, dp);
-    return port;
+    return scm_c_make_port (scm_dico_port_type,
+			    SCM_BUF0 | SCM_WRTNG, (scm_t_bits) dp);
 }
 
 #define DICO_PORT(x) ((struct _guile_dico_port *) SCM_STREAM (x))
 
-static SCM
-_dico_port_mark(SCM port)
-{
-    return SCM_BOOL_F;
-}
-
 static void
-_dico_port_flush(SCM port)
-{
-    struct _guile_dico_port *dp = DICO_PORT(port);
-    if (dp && dp->str)
-	dico_stream_flush(dp->str);
-}
-
-static int
 _dico_port_close(SCM port)
 {
     struct _guile_dico_port *dp = DICO_PORT(port);
 
-    if (dp) {
-	_dico_port_flush(port);
-	SCM_SETSTREAM(port, NULL);
-	scm_gc_free(dp, sizeof(struct _guile_dico_port), "dico-port");
-    }
-    return 0;
+    if (dp && dp->str)
+	dico_stream_flush(dp->str);
 }
 
-static scm_sizet
-_dico_port_free(SCM port)
-{
-    _dico_port_close(port);
-    return 0;
-}
-
-static int
-_dico_port_fill_input(SCM port)
-{
-    return EOF;
-}
-
-static void
-_dico_port_write(SCM port, const void *data, size_t size)
+static size_t
+_dico_port_write(SCM port, SCM src, size_t start, size_t count)
 {
     struct _guile_dico_port *dp = DICO_PORT(port);
-    dico_stream_write(dp->str, data, size);
+    dico_stream_write(dp->str, SCM_BYTEVECTOR_CONTENTS (src) + start,
+		      count);
+    return count;
 }
 
 static scm_t_off
@@ -616,22 +580,18 @@ _dico_port_print(SCM exp, SCM port, scm_print_state *pstate)
 }
 
 static void
-_guile_init_dico_port()
+_guile_init_dico_port(void)
 {
-    scm_tc16_dico_port = scm_make_port_type("dico-port",
-					    _dico_port_fill_input,
+    scm_dico_port_type = scm_make_port_type("dico-port",
+					    NULL,
 					    _dico_port_write);
-    scm_set_port_mark (scm_tc16_dico_port, _dico_port_mark);
-    scm_set_port_free (scm_tc16_dico_port, _dico_port_free);
-    scm_set_port_print (scm_tc16_dico_port, _dico_port_print);
-    scm_set_port_flush (scm_tc16_dico_port, _dico_port_flush);
-    scm_set_port_close (scm_tc16_dico_port, _dico_port_close);
-    scm_set_port_seek (scm_tc16_dico_port, _dico_port_seek);
+    scm_set_port_print (scm_dico_port_type, _dico_port_print);
+    scm_set_port_needs_close_on_gc (scm_dico_port_type, 1);
+    scm_set_port_close (scm_dico_port_type, _dico_port_close);
+    scm_set_port_seek (scm_dico_port_type, _dico_port_seek);
 }
-
-
 
-static scm_t_bits scm_tc16_dico_log_port;
+static scm_t_port_type *scm_dico_log_port_type;
 
 static SCM
 _make_dico_log_port(int level)
@@ -648,17 +608,15 @@ _dico_log_port_print(SCM exp, SCM port, scm_print_state *pstate)
 }
 
 static void
-_guile_init_dico_log_port()
+_guile_init_dico_log_port(void)
 {
-    scm_tc16_dico_log_port = scm_make_port_type("dico-log-port",
-						_dico_port_fill_input,
+    scm_dico_log_port_type = scm_make_port_type("dico-log-port",
+						NULL,
 						_dico_port_write);
-    scm_set_port_mark (scm_tc16_dico_log_port, _dico_port_mark);
-    scm_set_port_free (scm_tc16_dico_log_port, _dico_port_free);
-    scm_set_port_print (scm_tc16_dico_log_port, _dico_log_port_print);
-    scm_set_port_flush (scm_tc16_dico_log_port, _dico_port_flush);
-    scm_set_port_close (scm_tc16_dico_log_port, _dico_port_close);
-    scm_set_port_seek (scm_tc16_dico_log_port, _dico_port_seek);
+    scm_set_port_print (scm_dico_log_port_type, _dico_log_port_print);
+    scm_set_port_close (scm_dico_log_port_type, _dico_port_close);
+    scm_set_port_needs_close_on_gc (scm_dico_log_port_type, 1);
+    scm_set_port_seek (scm_dico_log_port_type, _dico_port_seek);
 }
 
 
@@ -688,9 +646,10 @@ enum guile_proc_ind {
     compare_count_proc,
     free_result_proc,
     result_headers_proc,
-};
+    db_mime_header_proc,
 
-#define MAX_PROC (result_headers_proc+1)
+    MAX_PROC
+};
 
 static char *guile_proc_name[] = {
     "open",
@@ -704,7 +663,8 @@ static char *guile_proc_name[] = {
     "result-count",
     "compare-count",
     "free-result",
-    "result-headers"
+    "result-headers",
+    "db-mime-header"
 };
 
 typedef SCM guile_vtab[MAX_PROC];
@@ -758,7 +718,7 @@ init_vtab(const char *init_fun, const char *dbname, guile_vtab vtab)
     if (guile_safe_exec(call_init_handler, &istr, &res))
 	return 1;
 
-    if (!scm_is_pair(res) && res != SCM_EOL) {
+    if (!scm_list_p(res) && res != SCM_EOL) {
 	str_rettype_error(init_fun);
 	return 1;
     }
@@ -767,7 +727,7 @@ init_vtab(const char *init_fun, const char *dbname, guile_vtab vtab)
 	char *ident;
 	SCM name, proc;
 	SCM car = SCM_CAR(res);
-	if (!scm_is_pair(res)
+	if (!scm_list_p(res)
 	    || !scm_is_string(name = SCM_CAR(car))
 	    || !scm_procedure_p(proc = SCM_CDR(car)))  {
 	    str_rettype_error(init_fun);
@@ -855,7 +815,7 @@ mod_init(int argc, char **argv)
 }
 
 static dico_handle_t
-mod_init_db(const char *dbname, int argc, char **argv)
+mod_init_db(const char *dbname, int argc, char **argv, void *extra)
 {
     struct _guile_database *db;
     int i;
@@ -990,7 +950,7 @@ scm_to_assoc(dico_assoc_list_t assoc, SCM scm)
 {
     dico_assoc_clear(assoc);
 
-    for (; scm != SCM_EOL && scm_is_pair(scm); scm = SCM_CDR(scm)) {
+    for (; scm != SCM_EOL && scm_list_p(scm); scm = SCM_CDR(scm)) {
 	SCM elt = SCM_CAR(scm);
 
 	if (!scm_is_pair(elt)) {
@@ -1058,9 +1018,9 @@ scm_to_langlist(SCM scm, SCM procsym)
     else if (scm_is_string(scm)) {
 	list = dico_list_create();
 	dico_list_append(list, scm_to_locale_string(scm));
-    } else if (scm_is_pair(scm)) {
+    } else if (scm_list_p(scm)) {
 	list = dico_list_create();
-	for (; scm != SCM_EOL && scm_is_pair(scm); scm = SCM_CDR(scm))
+	for (; scm != SCM_EOL && scm_list_p(scm); scm = SCM_CDR(scm))
 	    dico_list_append(list, scm_to_locale_string(SCM_CAR(scm)));
     } else
 	rettype_error(procsym);
@@ -1240,7 +1200,7 @@ mod_result_headers (dico_result_t rp, dico_assoc_list_t hdr)
 	if (guile_call_proc(&res, proc,
 			    scm_list_2(gres->result, assoc_to_scm(hdr))))
 	    return 1;
-	if (!scm_is_pair(res)) {
+	if (!scm_list_p(res)) {
 	    rettype_error(proc);
 	    return 1;
 	}
@@ -1249,22 +1209,30 @@ mod_result_headers (dico_result_t rp, dico_assoc_list_t hdr)
     return 0;
 }
 
+static char *
+mod_db_mime_header(dico_handle_t hp)
+{
+    struct _guile_database *db = (struct _guile_database *)hp;
+    return mod_get_text(db, db_mime_header_proc);
+}
+
 struct dico_database_module DICO_EXPORT(guile, module) = {
-    DICO_MODULE_VERSION,
-    DICO_CAPA_NONE,
-    mod_init,
-    mod_init_db,
-    mod_free_db,
-    mod_open,
-    mod_close,
-    mod_info,
-    mod_descr,
-    mod_lang,
-    mod_match,
-    mod_define,
-    mod_output_result,
-    mod_result_count,
-    mod_compare_count,
-    mod_free_result,
-    mod_result_headers
+    .dico_version = DICO_MODULE_VERSION,
+    .dico_capabilities = DICO_CAPA_NONE,
+    .dico_init = mod_init,
+    .dico_init_db = mod_init_db,
+    .dico_free_db = mod_free_db,
+    .dico_open = mod_open,
+    .dico_close = mod_close,
+    .dico_db_info = mod_info,
+    .dico_db_descr = mod_descr,
+    .dico_db_lang = mod_lang,
+    .dico_match = mod_match,
+    .dico_define = mod_define,
+    .dico_output_result = mod_output_result,
+    .dico_result_count = mod_result_count,
+    .dico_compare_count = mod_compare_count,
+    .dico_free_result = mod_free_result,
+    .dico_result_headers = mod_result_headers,
+    .dico_db_mime_header = mod_db_mime_header
 };
